/*
    bt866 - C O N E X A N T Bt866 video encoder driver version 0.0.1

    Copyright (C) 1999 Kazhiro Mogi <ee69011@isc.meiji.ac.jp>

  based on.
    
    saa7185 - Philips SAA7185B video encoder driver version 0.0.3

    Copyright (C) 1998 Dave Perks <dperks@ibm.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>

#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include <linux/i2c.h>
#include "video_encoder.h"

#define DEBUG(x)   x   /* Debug driver */   

/* ----------------------------------------------------------------------- */

struct bt866
{
   struct i2c_bus   *bus;
   int      addr;
   unsigned char   reg[128];

   int      norm;
   int      enable;
   int      bright;
   int      contrast;
   int      hue;
   int      sat;
};

#define   I2C_bt866        0x88

#define I2C_DELAY   10

/* ----------------------------------------------------------------------- */

static int bt866_write(struct bt866 *dev, unsigned char subaddr, unsigned char data)
{
   int ack;
   unsigned long flags;

   LOCK_I2C_BUS(dev->bus);

   i2c_start(dev->bus);
   i2c_sendbyte(dev->bus, dev->addr, I2C_DELAY);
   i2c_sendbyte(dev->bus, subaddr, I2C_DELAY);
   ack = i2c_sendbyte(dev->bus, data, I2C_DELAY);
   dev->reg[subaddr] = data;
   i2c_stop(dev->bus);
   UNLOCK_I2C_BUS(dev->bus);
   return ack;
}

static int bt866_write_block(struct bt866 *dev, unsigned const char *data, unsigned int len)
{
   int ack;
   unsigned subaddr;
   unsigned long flags;

   while (len > 1) {
      LOCK_I2C_BUS(dev->bus);
      i2c_start(dev->bus);
      i2c_sendbyte(dev->bus, dev->addr, I2C_DELAY);
      ack = i2c_sendbyte(dev->bus, (subaddr = *data++), I2C_DELAY);
      ack = i2c_sendbyte(dev->bus, (dev->reg[subaddr] = *data++), I2C_DELAY);
      len -= 2;
      while (len > 1 && *data == ++subaddr) {
         data++;
         ack = i2c_sendbyte(dev->bus, (dev->reg[subaddr] = *data++), I2C_DELAY);
         len -= 2;
      }
      i2c_stop(dev->bus);
      UNLOCK_I2C_BUS(dev->bus);
   }
   return ack;
}

/* ----------------------------------------------------------------------- */

static const unsigned char init_common[] = {

   /* OSDY8 - OSDY15, Used for 8 Segment Color Bar */
   0x70, 0xeb,   /* OSDY8,  White   */
   0x72, 0xa2,   /* OSDY9,  Yellow  */
   0x74, 0x83,   /* OSDY10, Cyan    */
   0x76, 0x70,   /* OSDY11, Green   */
   0x78, 0x54,   /* OSDY12, Magenta */
   0x7a, 0x41,   /* OSDY13, Red     */
   0x7c, 0x23,   /* OSDY14, Blue    */
   0x7e, 0x10,   /* OSDY15, Black   */

   /* OSDCr8 - OSDCr15        */
   0x90, 0x80,   /* OSDCr8    */
   0x92, 0x8e,   /* OSDCr9    */
   0x94, 0x2c,   /* OSDCr10   */
   0x96, 0x3a,   /* OSDCr11   */
   0x98, 0xc6,   /* OSDCr12   */
   0x9a, 0xd4,   /* OSDCr13   */
   0x9c, 0x72,   /* OSDCr14   */
   0x9e, 0x80,   /* OSDCr15   */

   /* OSDCb8 - OSDCb15        */
   0xb0, 0x8e,   /* OSDCb8    */
   0xb2, 0x2c,   /* OSDCb9    */
   0xb4, 0x9c,   /* OSDCb10   */
   0xb6, 0x48,   /* OSDCb11   */
   0xb8, 0xb8,   /* OSDCb12   */
   0xba, 0x64,   /* OSDCb13   */
   0xbc, 0xd4,   /* OSDCb14   */
   0xbe, 0x80,   /* OSDCb15   */

   /* Software Reset */
   0xC0, 0x00,   /* Normal Operation */

   /* Cr and Cb Scaling */
   0xc8, 0xc1,   /* CCIR 624 With Setup ?            */
   0xca, 0x89,   /* I don't know this value's detail */

   /* I/O Setup */
   0xcc, 0x20,   /* DAC     - ON,          (MSB)
                    YCSWAP  - OFF,
                    YC16    - 16 bit,
                    ESTATUS - Version,
                    ECGGATE - Normal,
                    OSDNUM  - Complement,
                    OSDEXT  - 8 Segment,
                    OSDBAR  - OSD Pin      (LSB)     */

  0xd6, 0x00,   /* HSYNC[7:0] Not Used in This Mode  */
  0xd8, 0x00,   /* VSYNC[7:0]                        */
  0xda, 0x01,   /* SYNC[2:0]  - 000b,      (MSB)
                   HSYNC[9:8] - 00b,
                   VSYNC[9:8] - 00b,
                   PORCH      - CCIR 601   (LSB)     */

  0xdc, 0x00,   /* SETUP     - Normal,     (MSB)
                   CBSWAP    - OFF,
                   SETMODE   - Autodetect,
                   MODE[4:3] - 00b,
                   MODE[2]   - NTSC,
                   MODE[1]   - 525 or 625 line,
                   MODE[0]   - CCIR 601    (LSB)     */
                 
  0xde, 0x07,   /* ECCF2   - Disable,      (MSB)             
                   ECCF1   - Disable,
                   DCHROMA - Normal,
                   F625L   - Normal,
                   SCRESET - Normal,
                   EVBI    - Never Added VBI,
                   EACTIVE - 1
                   ECLIP   - 1             (LSB)     */

};

static const unsigned char init_pal[] = {
   0xdc, 0x04,   /* MODE[2] - PAL  */
};

static const unsigned char init_ntsc[] = {
   0xdc, 0x00,   /* MODE[2] - NTSC */
};

static int bt866_attach(struct i2c_device *device)
{
   int i;
   struct bt866 * encoder;

   device->data = encoder = kmalloc(sizeof(struct bt866), GFP_KERNEL);
   if (encoder == NULL) {
      return -ENOMEM;
   }

   MOD_INC_USE_COUNT;

   memset(encoder, 0, sizeof(struct bt866));
   strcpy(device->name, "bt866");
   encoder->bus = device->bus;
   encoder->addr = device->addr;
   encoder->norm = VIDEO_MODE_NTSC;
   encoder->enable = 1;

   i = bt866_write_block(encoder, init_common, sizeof(init_common));
   if (i >= 0) {
      i = bt866_write_block(encoder, init_ntsc, sizeof(init_ntsc));
   }
   if (i < 0) {
      printk(KERN_ERR "%s_attach: init error %d\n", device->name, i);
   }

   return 0;
}


static int bt866_detach(struct i2c_device *device)
{
   kfree(device->data);
   MOD_DEC_USE_COUNT;
   return 0;
}

static int bt866_command(struct i2c_device *device, unsigned int cmd, void * arg)
{
   struct bt866 * encoder = device->data;

   switch (cmd) {

   case ENCODER_GET_CAPABILITIES:
      {
         struct video_encoder_capability *cap = arg;

         cap->flags
            = VIDEO_ENCODER_PAL
            | VIDEO_ENCODER_NTSC
            | VIDEO_ENCODER_CCIR;
         cap->inputs = 1;
         cap->outputs = 1;
      }
      break;

   case ENCODER_SET_NORM:
      {
         int * iarg = arg;

         switch (*iarg) {

         case VIDEO_MODE_NTSC:
            bt866_write_block(encoder, init_ntsc, sizeof(init_ntsc));
            break;

         case VIDEO_MODE_PAL:
            bt866_write_block(encoder, init_pal, sizeof(init_pal));
            break;

         case VIDEO_MODE_SECAM:
         default:
            return -EINVAL;

         }
         encoder->norm = *iarg;
      }
      break;

   case ENCODER_SET_INPUT:
      {
         int * iarg = arg;

#if 0
         /* not much choice of inputs */
         if (*iarg != 0) {
            return -EINVAL;
         }
#else
         /* RJ: *iarg = 0: input is from SA7111
                *iarg = 1: input is from ZR36060 */

         switch (*iarg) {

         case 0:
            /* Switch RTCE to 1 */
            bt866_write(encoder, 0xde, (encoder->reg[0xde] & 0xfb) | 0x04);
            break;

         case 1:
            /* Switch RTCE to 0 */
            bt866_write(encoder, 0xde, (encoder->reg[0xde] & 0xfb) | 0x00);
            break;

         default:
            return -EINVAL;

         }
#endif
      }
      break;

   case ENCODER_SET_OUTPUT:
      {
         int * iarg = arg;

         /* not much choice of outputs */
         if (*iarg != 0) {
            return -EINVAL;
         }
      }
      break;

   case ENCODER_ENABLE_OUTPUT:
      {
         int * iarg = arg;

         encoder->enable = !!*iarg;
         bt866_write(encoder, 0xcc, (encoder->reg[0xcc] & 0x7f) | (encoder->enable? 0x00: 0x80));
      }
      break;

   default:
      return -EINVAL;
   }

   return 0;
}

/* ----------------------------------------------------------------------- */

struct i2c_driver i2c_driver_bt866 = {
   "bt866",      /* name */
   I2C_DRIVERID_ENCODER,   /* ID */
   I2C_bt866, I2C_bt866+2,

   bt866_attach,
   bt866_detach,
   bt866_command
};

EXPORT_NO_SYMBOLS;

#ifdef MODULE
int init_module(void)
#else
int bt866_init(void)
#endif
{
   return i2c_register_driver(&i2c_driver_bt866);
}



#ifdef MODULE

void cleanup_module(void)
{
   i2c_unregister_driver(&i2c_driver_bt866);
}

#endif
