/*
    ks0127 - Samsung KS0127 video decoder driver version 0.0.1

	Copyright (C) 1999 Kazhiro Mogi <ee69011@isc.meiji.ac.jp>

  based on
	
    saa7111 - Philips SAA7111A video decoder driver version 0.0.3

    Copyright (C) 1998 Dave Perks <dperks@ibm.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>

#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include <linux/i2c.h>
#include "video_decoder.h"

#define DEBUG(x)       /* Debug driver */   

/* ----------------------------------------------------------------------- */

struct ks0127
{
   struct i2c_bus   *bus;
   int      addr;
   unsigned char   reg[128];

   int      norm;
   int      input;
   int      enable;
   int      bright;
   int      contrast;
   int      hue;
   int      sat;
};

#define   I2C_KS0127  0xd8

#define   I2C_DELAY   10

/* ----------------------------------------------------------------------- */

static int ks0127_write(struct ks0127 * dev, unsigned char subaddr, unsigned char data)
{
   int ack;
   unsigned long flags;

   LOCK_I2C_BUS(dev->bus);
   i2c_start(dev->bus);
   i2c_sendbyte(dev->bus, dev->addr, I2C_DELAY);
   i2c_sendbyte(dev->bus, subaddr, I2C_DELAY);
   ack = i2c_sendbyte(dev->bus, data, I2C_DELAY);
   dev->reg[subaddr] = data;
   i2c_stop(dev->bus);
   UNLOCK_I2C_BUS(dev->bus);
   return ack;
}

static int ks0127_write_block(struct ks0127 * dev, unsigned const char *data, unsigned int len)
{
   int ack;
   unsigned subaddr;
   unsigned long flags;

   ack = 0;
      
   while (len > 1) {
      LOCK_I2C_BUS(dev->bus);
      i2c_start(dev->bus);
      i2c_sendbyte(dev->bus, dev->addr, I2C_DELAY);
      ack = i2c_sendbyte(dev->bus, (subaddr = *data++), I2C_DELAY);
      ack = i2c_sendbyte(dev->bus, (dev->reg[subaddr] = *data++), I2C_DELAY);
      len -= 2;
      while (len > 1 && *data == ++subaddr) {
         data++;
         ack = i2c_sendbyte(dev->bus, (dev->reg[subaddr] = *data++), I2C_DELAY);
         len -= 2;
      }
      i2c_stop(dev->bus);
      UNLOCK_I2C_BUS(dev->bus);
   }
   return ack;
}

static int ks0127_read(struct ks0127 * dev, unsigned char subaddr)
{
   int data;
   unsigned long flags;

   LOCK_I2C_BUS(dev->bus);
   i2c_start(dev->bus);
   i2c_sendbyte(dev->bus, dev->addr, I2C_DELAY);
   i2c_sendbyte(dev->bus, subaddr, I2C_DELAY);
   i2c_start(dev->bus);
   i2c_sendbyte(dev->bus, (dev->addr | 1), I2C_DELAY);
   data = i2c_readbyte(dev->bus, 1);
   i2c_stop(dev->bus);
   UNLOCK_I2C_BUS(dev->bus);
   return data; 
}

/* ----------------------------------------------------------------------- */

static int ks0127_attach(struct i2c_device * device)
{
   int i;
   struct ks0127 * decoder;

   static const unsigned char init[] = {

	  /* Control Register A-D */
      0x01, 0x0c,      /* CCIR601, XTCL 24.576 MHz, Fast Loop, VSE 10/10.5 */
      0x02, 0x01,      /* AY1 Composite, VALIGN 0, AGC Run */
	  0x03, 0x40,      /* Enable TSTGPK, TSTGPK = 1 */
	  0x04, 0x00,      /* CMDD: KS0127 Defaults */

	  /* Horizontal Timing Signal Range */
	  0x05, 0x14,      /* HAVB: default + 20    */
	  0x06, 0x14,      /* HAVE: default + 20    */
	  0x07, 0x00,      /* HS1B: default         */
	  0x08, 0x00,      /* HS1E: default         */
	  0x09, 0x00,      /* HS2B: default         */
	  0x0a, 0x00,      /* HS2E: default         */
	  0x0b, 0x40,      /* AGC:  0x40            */
	  0x0c, 0x00,

	  /* Chroma Trap */
	  0x0f, 0x24,      /* Chroma trap, for Compsite Input */

	  /* Luma */
      0x10, 0x00,      /* CON   Contrast Control */
      0x11, 0x00,      /* BRT   Brightness Control */

	  /* Chroma Control */
	  0x12, 0x00,      /* KS0127 Defaults */
	  0x13, 0x00,      /* KS0127 Defaults */
	  0x14, 0x00,      /* KS0127 Defaults */

	  /* Chroma */
      0x15, 0x00,      /* SAT   Color Saturation Control */
      0x16, 0x00,      /* HUE   Hue Control */

	  /* Luma & Chroma Filter */
	  0x17, 0x00,      /*                                */
	  0x18, 0x12,      /* Now Testing HERE               */
	  0x19, 0x00,      /*                                */

	  /* HSCALE & VSCALE */
	  0x1a, 0x00,      /*                                */
	  0x1b, 0x00,      /* Now                            */
	  0x1c, 0x00,      /* Testing HERE                   */
	  0x1d, 0x00,      /*                                */
	  
	  /* Output Control */
	  0x1e, 0x20,      /* OENC 2 */

	  /* VBI Decoder Control */
	  0x20, 0x00,      /* KS0127 Defaults */

	  /* UV Offset Adjustment */
	  0x29, 0x00,      /* ? KS0127 Defaults */
	  0x2a, 0x00,      /* ? */
	  
	  /* Chroma Tracking Control Register */
	  0x2f, 0x0f,      /* Chroma Gain and Freq tracking Faster */

	  /* Timing Code Polarity Control */
	  0x30, 0x00,      /* All Signal Polarities are High */
	  
	  /* Reference Code Insertion Control */
	  0x31, 0x6c,      /* Insert VS and HS timing code */

	  /* Invalid YUV Code */
	  0x32, 0x80,      /* INVALY */
	  0x33, 0x80,      /* INVALU */
	  0x34, 0x80,      /* INVALV */

	  /* Unused YUV Code - HAV is inactive YUV code */
	  0x35, 0xff,      /* UNUSEY */
	  0x36, 0x80,      /* UNUSEU */
	  0x37, 0x80,      /* UNUSEV */

	  /* Commnad Register E-F */
	  0x3d, 0x00,      /* KS0127 Default */
	  0x3f, 0x00,      /* KS0127 Default */
   };
   

   device->data = decoder = kmalloc(sizeof(struct ks0127), GFP_KERNEL);
   if (decoder == NULL) {
      return -ENOMEM;
   }

   MOD_INC_USE_COUNT;

   memset(decoder, 0, sizeof(struct ks0127));
   strcpy(device->name, "ks0127");
   decoder->bus = device->bus;
   decoder->addr = device->addr;
   decoder->norm = VIDEO_MODE_AUTO;
   decoder->input = 0;
   decoder->enable = 1;
   decoder->bright = 32768;
   decoder->contrast = 32768;
   decoder->hue = 32768;
   decoder->sat = 32768;

   i = ks0127_write_block(decoder, init, sizeof(init));
   if (i < 0) {
      printk(KERN_ERR "%s_attach: init status %d\n", device->name, i);
   } else {
      printk(KERN_INFO "%s_attach: chip version %x\n", device->name, ks0127_read(decoder, 0x00));
   }
   return 0;
}


static int ks0127_detach(struct i2c_device * device)
{
   kfree(device->data);
   MOD_DEC_USE_COUNT;
   return 0;
}

static int ks0127_command(struct i2c_device * device, unsigned int cmd, void * arg)
{
   struct ks0127 * decoder = device->data;

   switch (cmd) {

#if defined(DECODER_DUMP)
   case DECODER_DUMP:
      {
         int i;

         for (i = 0; i < 32; i += 16) {
            int j;

            printk("KERN_DEBUG %s: %03x", device->name, i);
            for (j = 0; j < 16; ++j) {
               printk(" %02x", ks0127_read(decoder, i + j));
            }
            printk("\n");
         }
      }
      break;
#endif /* defined(DECODER_DUMP) */

   case DECODER_GET_CAPABILITIES:
      {
         struct video_decoder_capability *cap = arg;

         cap->flags
            = VIDEO_DECODER_PAL
            | VIDEO_DECODER_NTSC
            | VIDEO_DECODER_AUTO
            | VIDEO_DECODER_CCIR;
         cap->inputs = 2;
         cap->outputs = 1;
      }
      break;

   case DECODER_GET_STATUS:
      {
         int * iarg = arg;
         int status;
         int res;

         status = ks0127_read(decoder, 0x00);
		 
         res = (status << 8);
         if ((status & 0x04) != 0) {
            res |= DECODER_STATUS_GOOD;
         }
         switch (decoder->norm) {
         case VIDEO_MODE_NTSC:
            res |= DECODER_STATUS_NTSC;
            break;
         case VIDEO_MODE_PAL:
            res |= DECODER_STATUS_PAL;
            break;
         default:
         case VIDEO_MODE_AUTO:
			 if ((status & 0x08) != 0) {
				 res |= DECODER_STATUS_PAL;
			 } else {
				 res |= DECODER_STATUS_NTSC;
			 }
            break;
         }
         if ((status & 0x04) != 0) {
            res |= DECODER_STATUS_COLOR;
         }
         *iarg = res;
      }
      break;

   case DECODER_SET_NORM:
      {
         int * iarg = arg;

         switch (*iarg) {

         case VIDEO_MODE_NTSC:
            ks0127_write(decoder, 0x12, (decoder->reg[0x12] & 0x9f) | 0x00);
            break;

         case VIDEO_MODE_PAL:
            ks0127_write(decoder, 0x12, (decoder->reg[0x12] & 0x9f) | 0x60);
            break;

         case VIDEO_MODE_AUTO:
            ks0127_write(decoder, 0x12, (decoder->reg[0x12] & 0x9f) | 0x00);
            break;

         default:
            return -EINVAL;

         }
         decoder->norm = *iarg;
      }
      break;

   case DECODER_SET_INPUT:
      {
         int * iarg = arg;

         if (*iarg < 0 || *iarg > 1) {
            return -EINVAL;
         }

         if (decoder->input != *iarg) {
            decoder->input = *iarg;
			if(decoder->input == 0){ /* Composite Input */
				ks0127_write(decoder, 0x02, (decoder->reg[0x02] & 0xf0) | 0x01);
				ks0127_write(decoder, 0x0f, (decoder->reg[0x0f] & 0xfb) | 0x04);
				ks0127_write(decoder, 0x12, (decoder->reg[0x02] & 0xfc) | 0x00);
			}else{ /* S-Video Input */
				ks0127_write(decoder, 0x02, (decoder->reg[0x02] & 0xf0) | 0x08);
				ks0127_write(decoder, 0x0f, (decoder->reg[0x0f] & 0xfb) | 0x00);
				ks0127_write(decoder, 0x12, (decoder->reg[0x02] & 0xfc) | 0x02);
			}
         }
      }
      break;

   case DECODER_SET_OUTPUT:
      {
         int * iarg = arg;

         /* not much choice of outputs */
         if (*iarg != 0) {
            return -EINVAL;
         }
      }
      break;

   case DECODER_ENABLE_OUTPUT:
      {
         int * iarg = arg;
         int enable = (*iarg != 0);

         if (decoder->enable != enable) {
            decoder->enable = enable;

// RJ: If output should be disabled (for playing videos), we also need a open PLL.
//     The input is set to 0 (where no input source is connected), although this
//     is not necessary.
//
//     If output should be enabled, we have to reverse the above.

            if(decoder->enable)
            {
               ks0127_write(decoder, 0x01, (decoder->reg[0x01] & 0x7f) | 0x00);
               ks0127_write(decoder, 0x03, (decoder->reg[0x03] & 0x7f) | 0x00);
            }
            else
            {
               ks0127_write(decoder, 0x01, (decoder->reg[0x01] & 0x7f) | 0x80);
               ks0127_write(decoder, 0x03, (decoder->reg[0x03] & 0x7f) | 0x80);
            }
         }
      }
      break;

   case DECODER_SET_PICTURE:
      {
         struct video_picture *pic = arg;

         if (decoder->bright != pic->brightness) {
            /* We want -128 to 127 we get 0-65535 */
            decoder->bright = pic->brightness;
            ks0127_write(decoder, 0x11, (decoder->bright - 32768) >> 8);
         }
         if (decoder->contrast != pic->contrast) {
            /* We want -128 to 127 we get 0-65535 */
            decoder->contrast = pic->contrast;
            ks0127_write(decoder, 0x10, (decoder->contrast - 32768) >> 8);
         }
         if (decoder->sat != pic->colour) {
            /* We want 0 to 255 we get 0-65535 */
            decoder->sat = pic->colour;
            ks0127_write(decoder, 0x15, (decoder->sat - 32768) >> 8);
         }
         if (decoder->hue != pic->hue) {
            /* We want 0 to 255 we get 0-65535 */
            decoder->hue = pic->hue;
            ks0127_write(decoder, 0x16, (decoder->hue - 32768) >> 8);
         }
      }
      break;

   default:
      return -EINVAL;
   }

   return 0;
}

/* ----------------------------------------------------------------------- */

struct i2c_driver i2c_driver_ks0127 = {
   "ks0127",      /* name */
   I2C_DRIVERID_DECODER,   /* ID */
   I2C_KS0127, I2C_KS0127+6,

   ks0127_attach,
   ks0127_detach,
   ks0127_command
};

EXPORT_NO_SYMBOLS;

#ifdef MODULE
int init_module(void)
#else
int ks0127_init(void)
#endif
{
   return i2c_register_driver(&i2c_driver_ks0127);
}



#ifdef MODULE

void cleanup_module(void)
{
   i2c_unregister_driver(&i2c_driver_ks0127);
}

#endif
